try:
    import cStringIO as StringIO
except ImportError:
    import StringIO

from twisted.web import client
from twisted.internet import reactor
from twisted.internet.error import TimeoutError
from twisted.internet.protocol import ReconnectingClientFactory

import feedparser

from utils import log, stripper
from data import Message

FIRST_CHOICE = -1
CONTENT_INDEX = 1

def extractFeedData(feed, entry):
    '''

    '''
    if entry.has_key('title'):
        title = stripper(entry.title)
    else:
        title = ''
    if entry.has_key('description'):
        description = stripper(entry.description)
        description = (len(description), description)
    else:
        description = (0,'')
    if entry.has_key('summary'):
        summary = stripper(entry.summary)
        summary = (len(summary), summary)
    else:
        summary = (0,'')
    if entry.has_key('content'):
        content = stripper(entry.content[0]['value'])
        content = (len(content), content)
    else:
        content = (0,'')

    choice = [description, summary, content]
    choice.sort()
    content = choice[FIRST_CHOICE][CONTENT_INDEX]

    if content:
        rank = int(feed['rank'])
        order = int(feed['order'])
        try:
            link = entry.content[0]['base']
        except:
            link = feed['link']

        return Message(rank=rank, title=title, content=content,
            link=link, order=order)

class ReconnectingHTTPClientFactory(ReconnectingClientFactory,
                                    client.HTTPClientFactory):
    protocol = client.HTTPPageGetter
    url = None
    scheme = None
    host = ''
    port = None
    path = None   
    maxDelay = 60

class FeedEntry(object):

    def __init__(self, feedData, pageTimeout=360, connectionTimeout=120):
        # feedData is a data.Message object
        self.metadata = feedData
        self.connectionTimeout = connectionTimeout
        self.errorCount = 0
        self.errorList = []
        self.data = {}
        self.agent = "Mozilla/5.0 (Macintosh; U; PPC Mac OS X Mach-O; " + \
            "en-US; rv:1.7.10) Gecko/20050716 Firefox/1.0.6"

        d = self.getPage(self.metadata.link, timeout=pageTimeout,
            agent=self.agent)
        d.addErrback(self.gotTimeout, 'getting', d)
        d.addErrback(self.gotError, 'getting')

        # Parse the feed and if there's some errors call 
        # self.gotError
        d.addCallback(self.parseFeed)
        d.addErrback(self.gotError, 'parsing')
        
        # When you get the raw structure you can work on it to 
        # format in the best way you can think of.  For any error 
        # call self.gotError.
        d.addCallback(self.workOnPage)
        d.addErrback(self.gotError, 'working on page')

        self.deferred = d

    def getPage(self, url, *args, **kwargs):
        scheme, host, port, path = client._parse(url)
        factory = ReconnectingHTTPClientFactory(url, *args, **kwargs)
        if scheme == 'https':
            from twisted.internet import ssl
            if contextFactory is None:
                contextFactory = ssl.ClientContextFactory()
            reactor.connectSSL(host, port, factory, contextFactory,
                timeout=self.connectionTimeout)
        else:
            self.connector = reactor.connectTCP(host, port, factory,
                timeout=self.connectionTimeout)
        return factory.deferred

    def gotTimeout(self, failure, comment, deferred):
        failure.trap(TimeoutError)
        msg = 'While %s feed %s (No. %s): %s' % (comment,
            self.metadata.link, self.metadata.rank,
            failure.getErrorMessage())
        log.error(msg)
        log.debug(str(self.connector))
        log.debug(str(self.connector.factory))
        log.debug("Connector timeout: %s" % str(self.connector.timeout))

    def gotError(self, failure, comment):
        # An Error as occurred, print traceback infos and go on
        msg = 'While %s feed %s (No. %s): %s' % (comment,
            self.metadata.link, self.metadata.rank,
            failure.getErrorMessage())
        log.error(msg)
        log.debug(failure.getTraceback())
        self.errorCount += 1
        self.errorList.append(failure)
        
    def parseFeed(self, page):
        if page:
            log.debug("Page type: %s" % type(page))
            log.info("Getting data from source #%s: '%s'..." % (
                self.metadata.rank, self.metadata.link))
            try:
                parsed = feedparser.parse(StringIO.StringIO(page))
            except TypeError:
                parsed = feedparser.parse(StringIO.StringIO(str(page)))
            except Exception, e:
                log.error("Failed to parse feed from source #%s: '%s'..." % (
                    self.metadata.rank, self.metadata.link))
                log.info(e)
                return 
            return parsed
   
    def workOnPage(self, parsedFeed):
        '''
        Feed is the webXcreta feed metadata
        '''
        log.debug("Finished retrieving %s" % self.metadata.link)
        try:
            msg = "Feed Version:", parsedFeed.get('version', 'Unknown')
            log.debug(msg)
            chan = parsedFeed.get('channel', None)
            if chan:
                log.debug(chan.get('title', ''))
            try:
                entry = parsedFeed.entries[0]
                self.data = extractFeedData(self.metadata,
                    entry)
            except IndexError:
                log.warning("No entry for %s." % self.metadata.link)
                if parsedFeed.has_key('status'):
                    log.debug("Status: %s" % parsedFeed.status)
            return self.data

        except AttributeError:
            # this means that there's no parsedFeed
            log.warning("No feedparser instance for %s" % (
                self.metadata.link))


