import email
import mailbox
import cPickle as pickle
from email.Errors import MessageParseError

from twisted.internet import defer

import feedparser

import feeds
from utils import log, extractEmailData, recoverableError
from feeds import extractFeedData

FIRST_CHOICE = -1

class BaseHarvester(object):

    def __init__(self):
        self.data = []
        self.highestRank = 1
        self.titles = {}
        self.entries = {}

    def harvest(self):
        raise Exception, NotImplemented

    def setHighRank(self):
        ranks = [i['rank'] for i in self.data]
        ranks.sort()
        log.debug("Sorted ranks: %s" % ranks)
        self.highestRank = ranks[FIRST_CHOICE]
        log.debug("Set highest rank to '%s'." % self.highestRank)

    def getHighRank(self):
        return self.highestRank

    def getData(self):
        return self.data

class FakeHarvester(BaseHarvester):
    pass
FakeHarvest = FakeHarvester

class SynchHarvester(BaseHarvester):

    def __init__(self, feeds_list):
        super(SynchHarvester, self).__init__()
        if not feeds_list:
            raise Exception, "Cannot harvest when there are no feeds."
        for feed in feeds_list:
            self.harvest(feed)
        self.setHighRank()

    def harvest(self, feed):
        source = feed['link']
        log.info("Getting data from source #%i: '%s'..." % (feed['rank'], source))
        try:
            p = feedparser.parse(source)
        except LookupError, e:
            msg = "Encountered a problem with #%i, '%s' ..."  % (
                feed['rank'], source)
            recoverableError(msg, e)
            return
        except Exception, e:
            msg = "Error with #%i, '%s' ..."  % (
                feed['rank'], source)
            recoverableError(msg, e)
            return
        try:
            entry = p.entries[0]
            data = extractFeedData(feed, entry)
            if data:
                self.data.append(data)
                log.debug("Bytes read: %s" % len(self.data))
        except IndexError:
            log.warning("No entry for %s." % feed['link'])
            if p.has_key('status'):
                log.debug("Status: %s" % p.status)
            if p.has_key('bozo_exception'):
                try:
                    log.debug(p.bozo_exception.reason.args[1])
                except:
                    pass
        except Exception, e:
            msg = "Encountered a problem with #%i, '%s' ..."  % (
                feed['rank'], source)
            recoverableError(msg, e)
        if not self.data:
            error = "Harvester returned no data for #%i, '%s'." % (
                feed['rank'], source)
            recoverableError(None, error)

class AsynchHarvester(BaseHarvester):

    def __init__(self):
        super(AsynchHarvester, self).__init__()
        self.feedCounter = 0
        self.deferred = defer.Deferred()
        self.totalFeeds = 0

    def setupFeedDeferred(self, feed):
        self.feedCounter += 1
        name = "deferred #%s" % self.feedCounter
        fe = feeds.FeedEntry(feed)
        d = fe.deferred
        d.addCallback(self.updateData)
        return d

    def updateData(self, data):
        if data:
            self.data.append(data)
            log.debug("New data length: %s" % len(self.data))

    def handleAllGroups(self, data):
        log.info("All groups completed!")
        self.setHighRank()
        log.debug("Highest rank: %s" % self.getHighRank())
        return self

    def handleFailure(self, failure):
        log.error("We got an exception: %s" % (failure.getTraceback(),))

class GroupHarvester(AsynchHarvester):

    def __init__(self):
        super(GroupHarvester, self).__init__()
        self.group_counter = 0
        self.groups = iter([])
        self.setTimeout(500)
        self.setHostsPerGroup(3)
        self.group_deferreds = []

    def setTimeout(self, timeout):
        self.timeout = timeout

    def setHostsPerGroup(self, host_count):
        self.hosts_per_group = host_count

    def harvest(self, feeds):
        self.totalFeeds = len(feeds)
        self.groups = self.splitLoad(feeds, self.hosts_per_group)
        # iterate through each group
        host_group = self.groups.next()
        dl = self.setupGroupDeferred(host_group)
        dl.addCallback(self.getNextGroup)

    def setupGroupDeferred(self, group):
        self.group_counter += 1
        # iterate through each feed in the group
        deferreds = [ self.setupFeedDeferred(feed) for feed in group ]
        # set up this group as a deferred list
        dl = defer.DeferredList(deferreds)
        dl.addCallback(self.handleGroup, self.group_counter)
        dl.addErrback(self.handleFailure)
        return dl

    def getNextGroup(self, null):
        group = self.groups.next()
        dl = self.setupGroupDeferred(group)
        dl.addCallback(self.getNextGroup)
        dl.addErrback(self.noMoreGroups)

    def getDeferred(self):
        return self.deferred

    def splitLoad(self, host_list, per_group):
        group_count, remainder = divmod(len(host_list), per_group)
        if remainder: group_count += 1
        for i in xrange(group_count):
            yield host_list[i*per_group:i*per_group+per_group]

    def handleGroup(self, data, count):
        log.info("Finished with Group %s" % count)

    def noMoreGroups(self, failure):
        failure.trap(StopIteration)
        self.setHighRank()
        log.info("Highest rank: %s" % self.getHighRank())
        log.info("Ratio of successful gets to total feeds: %s/%s" % (
            len(self.data), self.totalFeeds))
        self.deferred.callback(self)

class MboxHarvester(BaseHarvester):

    def __init__(self, mboxfile):
        super(MboxHarvester, self).__init__()
        self.mboxfile = mboxfile
        self.mbox = open(mboxfile)
        self.harvest()
        self.highestRank = 1

    def msgFactory(self, mbox):
        try:
            return email.message_from_file(mbox)
        except email.Errors.MessageParseError:
            # Don't return None since that will
            # stop the mailbox iteratorv
            return ''

    def harvest(self):
        log.info('Havesting data from mbox: %s' % self.mboxfile)
        mbox = mailbox.PortableUnixMailbox(self.mbox, self.msgFactory)
        for msg in mbox:
            data = extractEmailData(msg)
            if data:
                self.data.append(data)


